/* $AoiMoe: gesture_conf.hh,v 1.3 2003/02/06 14:31:00 tshiozak Exp $ */

/*-
 * Copyright (c)2003 Takuya SHIOZAKI,
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef GESTURE_CONF_HH_
#define GESTURE_CONF_HH_

#ifdef __GNUC__
#pragma interface "gesture_conf.hh"
#endif


// ----------------------------------------------------------------------
// macros
//

///
#define DEF_HANDLE(name)				\
public:							\
	typedef name *Handle // XXX: should be shared_ptr


// ----------------------------------------------------------------------
// exceptions
//

class Base_E : public std::exception
{
public:
	~Base_E() { }
	const char *what() const { return ("Base exception"); }
};

class UnknownKeysym_E : public Base_E
{
	std::string m_message_, m_keysym_;
public:
	UnknownKeysym_E(std::string const &sym)
		: m_keysym_(sym)
	{
		m_message_ = "Unknown keysym: ";
		m_message_ += m_keysym_;
	}
	const char *what() const
	{
		return (m_message_.c_str());
	}
};

class NotImpl_E : public Base_E
{
public:
	NotImpl_E() { }
	const char *what() const { return ("Not implemented"); }
};

class NoEntry_E : public Base_E
{
public:
	NoEntry_E() { }
	const char *what() const { return ("No entry"); }
};


// ----------------------------------------------------------------------
// gesture configuration container classes
//


/// keycode holder.
class KeyCode_C
{
	DEF_HANDLE(KeyCode_C);
private:
	int m_keycode_;
public:
	explicit KeyCode_C(int keycode) : m_keycode_(keycode) { }
	int GetKeyCode() const { return (m_keycode_); }
};


/// key stroke holder.
class Stroke_C
{
	DEF_HANDLE(Stroke_C);
private:
	list<KeyCode_C::Handle> m_buffer_;
	template <class GenKey_>
	class GenKeyWrapper_
	{
		GenKey_ m_genkey_;
		bool m_ispress_;
	public:
		GenKeyWrapper_(GenKey_ const &genkey, bool ispress)
			: m_genkey_(genkey), m_ispress_(ispress)
		{
		}
		void operator () (KeyCode_C::Handle ptr)
		{
			m_genkey_(ptr->GetKeyCode(), m_ispress_);
		}
	};
public:
	Stroke_C() { }
	~Stroke_C() { }
	void PushKeyCode(int key)
	{
		m_buffer_.push_back(new KeyCode_C(key));
	}
	template <class GenKey_>
	void GenerateKey(GenKey_ const &genkey)
	{
		for_each(m_buffer_.begin(), m_buffer_.end(),
			 GenKeyWrapper_<GenKey_>(genkey, true));
		for_each(m_buffer_.rbegin(), m_buffer_.rend(),
			 GenKeyWrapper_<GenKey_>(genkey, false));
	}
};


/// key sequece holder.
class Sequence_C
{
	DEF_HANDLE(Sequence_C);
private:
	list<Stroke_C::Handle> m_buffer_;
	template <class GenKey_>
	class GenKeyWrapper_
	{
		GenKey_ m_genkey_;
	public:
		GenKeyWrapper_(GenKey_ const &genkey) : m_genkey_(genkey) { }
		void operator () (Stroke_C::Handle ptr)
		{
			ptr->GenerateKey(m_genkey_);
		}
	};
public:
	Sequence_C() { }
	~Sequence_C() { }
	void PushStroke(Stroke_C::Handle stroke)
	{
		m_buffer_.push_back(stroke);
	}
	template <class GenKey_>
	void GenerateKey(GenKey_ const &genkey)
	{
		for_each(m_buffer_.begin(), m_buffer_.end(),
			 GenKeyWrapper_<GenKey_>(genkey));
	}
};


/// motion -> key sequence binder.
class Motion_C
{
	DEF_HANDLE(Motion_C);
private:
	std::string m_motion_;
	std::string m_title_;
	Sequence_C::Handle m_sequence_;
public:
	Motion_C(std::string const &m, std::string const &t,
		 Sequence_C::Handle s)
		: m_motion_(m), m_title_(t), m_sequence_(s)
	{
	}
	std::string const &GetMotion() const { return (m_motion_); }
	std::string const &GetTitle() const { return (m_title_); }
	Sequence_C::Handle const &GetSequence() const { return (m_sequence_); }
	~Motion_C() { }
};


/// motion set holder.
class MotionSet_C
{
	DEF_HANDLE(MotionSet_C);
private:
	std::map<std::string, Motion_C::Handle> m_motions_;
public:
	MotionSet_C() { }
	~MotionSet_C() { }
	void PushMotion(std::string const &m, std::string const &t,
			Sequence_C::Handle s)
	{
		Motion_C::Handle h = new Motion_C(m, t, s);
		m_motions_[m] = h;
	}
	Motion_C::Handle GetMotion(std::string const &m) const
	{
		std::map<std::string, Motion_C::Handle>::const_iterator iter;

		iter = m_motions_.find(m);
		if (iter == m_motions_.end())
			return (NULL);

		return (iter->second);
	}
	void ImportMotionSet(MotionSet_C::Handle src)
	{
		std::map<std::string, Motion_C::Handle>::const_iterator i;

		for (i=src->m_motions_.begin(); i!=src->m_motions_.end(); i++)
			m_motions_[i->first] = i->second;
	}
};


/// target application -> motion set binder.
class Target_C
{
	DEF_HANDLE(Target_C);
private:
	std::string m_name_;
	std::string m_title_;
	MotionSet_C::Handle m_motions_;
public:
	Target_C(std::string const &n, std::string const &t,
		 MotionSet_C::Handle ml)
		: m_name_(n), m_title_(t), m_motions_(ml)
	{
	}
	~Target_C() { }
	std::string const &GetAppName() const { return (m_name_); }
	std::string const &GetTitle() const { return (m_title_); }
	MotionSet_C::Handle GetMotionSet() const { return (m_motions_); }
};

/// target pool.
class TargetPool_C
{
	DEF_HANDLE(TargetPool_C);
private:
	list<Target_C::Handle> m_targets_;
public:
	TargetPool_C() { }
	~TargetPool_C() { }
	void
	PushTarget(std::string const &n, std::string const &t,
		   MotionSet_C::Handle ml)
	{
		m_targets_.push_back(new Target_C(n, t, ml));
	}
	template <class Pred_>
	Target_C::Handle Lookup(Pred_ const &pred) const
	{
		list<Target_C::Handle>::const_iterator iter;

		iter = find_if(m_targets_.begin(), m_targets_.end(), pred);
		if (iter == m_targets_.end())
			throw NoEntry_E();

		return (*iter);
	}
};


/// named motion set pool.
class MotionSetPool_C
{
	DEF_HANDLE(MotionSetPool_C);
private:
	std::map<std::string, MotionSet_C::Handle> m_list_;
public:
	MotionSetPool_C() { }
	~MotionSetPool_C() { }
	MotionSet_C::Handle Lookup(std::string const &name) const
	{
		std::map<std::string, MotionSet_C::Handle>::const_iterator i;

		i = m_list_.find(name);
		if (i == m_list_.end())
			throw NoEntry_E();

		return (i->second);
	}
	void PushMotionSet(std::string const &name, MotionSet_C::Handle ml)
	{
		m_list_[name] = ml;
	}
};


/// keysym -> keycode converter interface.
class KeysymConverter_I
{
	DEF_HANDLE(KeysymConverter_I);
public:
	virtual ~KeysymConverter_I() = 0;
	virtual int operator () (std::string const &) = 0;
};

/// error message procedure interface.
class ErrorMessageProc_I
{
	DEF_HANDLE(ErrorMessageProc_I);
public:
	virtual ~ErrorMessageProc_I() = 0;
	virtual void operator () (int,
				  std::string const &,
				  std::string const &) = 0;
};

/// details of parse manager.
namespace ParseManager_D_
{

class Impl_I
{
	DEF_HANDLE(Impl_I);
public:
	virtual ~Impl_I() = 0;
	virtual TargetPool_C::Handle GetTargetPool() const = 0;
	virtual void Parse(std::string const &) = 0;
	//
	static Impl_I *create(KeysymConverter_I::Handle,
			      ErrorMessageProc_I::Handle);
};

} // namespace ParseManager_D_



/// parse manager class.
class ParseManager_C
{
	DEF_HANDLE(ParseManager_C);
private:
	ParseManager_D_::Impl_I::Handle m_pimpl_;
public:
	ParseManager_C(KeysymConverter_I::Handle conv,
		       ErrorMessageProc_I::Handle err)
		: m_pimpl_(ParseManager_D_::Impl_I::create(conv, err))
	{
	}
	~ParseManager_C() { }
	//
	TargetPool_C::Handle GetTargetPool() const
	{
		return (m_pimpl_->GetTargetPool());
	}
	void Parse(std::string const &gesture)
	{
		return (m_pimpl_->Parse(gesture));
	}
};

#endif
